# -*- coding: utf-8 -*-
import os

from psycopg2cffi import compat

import django
import django.test.utils


os.environ.setdefault("DJANGO_SETTINGS_MODULE", "tests.settings")

compat.register()

django.setup()



# -*- coding: utf-8 -*-

import coreschema

from django.test import RequestFactory

from rest_framework.settings import api_settings
from rest_framework.test import APISimpleTestCase

from rest_witchcraft.filters import SearchFilter
from rest_witchcraft.serializers import ModelSerializer
from rest_witchcraft.viewsets import ModelViewSet

from .models import Owner, session


class OwnerSerializer(ModelSerializer):
    class Meta:
        model = Owner
        session = session
        fields = "first_name", "last_name"


class OwnerViewSet(ModelViewSet):
    serializer_class = OwnerSerializer
    session = session
    queryset = property(lambda self: session.query(Owner))


class TestSearchFilters(APISimpleTestCase):
    factory = RequestFactory()

    def setUp(self):
        super().setUp()
        self.viewset_class = OwnerViewSet
        self.filter = SearchFilter()

        self.owner1 = Owner(id=1, first_name="Joe", last_name="Smith")
        self.owner2 = Owner(id=2, first_name="Jon", last_name="Snow")

        self.owners = [self.owner1, self.owner2]
        session.add_all(self.owners)
        session.flush()

    def tearDown(self):
        session.rollback()
        super().tearDown()

    def test_search_field(self):
        viewset = self.viewset_class()
        viewset.action_map = {"get": "list"}

        schema = self.filter.get_schema_fields(viewset)

        self.assertEqual(len(schema), 1)
        field = schema[0]

        self.assertEqual(field.name, api_settings.SEARCH_PARAM)
        self.assertFalse(field.required)
        self.assertEqual(field.location, "query")
        self.assertIsInstance(field.schema, coreschema.String)

        schema_ops = self.filter.get_schema_operation_parameters(viewset)
        self.assertDictEqual(
            {"test": schema_ops},
            {
                "test": [
                    {
                        "name": "search",
                        "required": False,
                        "in": "query",
                        "description": "A search term.",
                        "schema": {"type": "string"},
                    }
                ]
            },
        )

    def test_to_html(self):
        viewset = self.viewset_class()
        viewset.action_map = {"get": "list"}
        request = viewset.initialize_request(self.factory.get("/", {api_settings.SEARCH_PARAM: "jo"}))

        html = self.filter.to_html(request, viewset.get_queryset(), viewset)

        self.assertInHTML("<h2>Search</h2>", html)

        del viewset.__class__.search_fields
        html = self.filter.to_html(request, viewset.get_queryset(), viewset)
        self.assertEqual(html, "")

    def test_icontains(self):
        self.viewset_class.search_fields = ["first_name", "last_name"]

        viewset = self.viewset_class()
        viewset.action_map = {"get": "list"}
        request = viewset.initialize_request(self.factory.get("/"))

        query = self.filter.filter_queryset(request, viewset.get_queryset(), viewset)

        self.assertEqual(set(query.all()), set(self.owners))

        request = viewset.initialize_request(self.factory.get("/", {api_settings.SEARCH_PARAM: "jon"}))
        query = self.filter.filter_queryset(request, viewset.get_queryset(), viewset)

        self.assertEqual(set(query.all()), {self.owner2})

    def test_istartswith(self):
        self.viewset_class.search_fields = ["^first_name", "^last_name"]
        viewset = self.viewset_class()
        viewset.action_map = {"get": "list"}

        request = viewset.initialize_request(self.factory.get("/", {api_settings.SEARCH_PARAM: "jo"}))
        query = self.filter.filter_queryset(request, viewset.get_queryset(), viewset)
        self.assertEqual(set(query.all()), set(self.owners))

        request = viewset.initialize_request(self.factory.get("/", {api_settings.SEARCH_PARAM: "sno"}))
        query = self.filter.filter_queryset(request, viewset.get_queryset(), viewset)
        self.assertEqual(set(query.all()), {self.owner2})

    def test_iexact(self):
        self.viewset_class.search_fields = ["=first_name", "=last_name"]
        viewset = self.viewset_class()
        viewset.action_map = {"get": "list"}

        request = viewset.initialize_request(self.factory.get("/", {api_settings.SEARCH_PARAM: "sno"}))
        query = self.filter.filter_queryset(request, viewset.get_queryset(), viewset)
        self.assertEqual(set(query.all()), set())

        request = viewset.initialize_request(self.factory.get("/", {api_settings.SEARCH_PARAM: "snow"}))
        query = self.filter.filter_queryset(request, viewset.get_queryset(), viewset)
        self.assertEqual(set(query.all()), {self.owner2})

    def test_exact(self):
        self.viewset_class.search_fields = ["@first_name", "@last_name"]
        viewset = self.viewset_class()
        viewset.action_map = {"get": "list"}

        request = viewset.initialize_request(self.factory.get("/", {api_settings.SEARCH_PARAM: "snow"}))
        query = self.filter.filter_queryset(request, viewset.get_queryset(), viewset)
        self.assertEqual(set(query.all()), set())

        request = viewset.initialize_request(self.factory.get("/", {api_settings.SEARCH_PARAM: "Snow"}))
        query = self.filter.filter_queryset(request, viewset.get_queryset(), viewset)
        self.assertEqual(set(query.all()), {self.owner2})



# -*- coding: utf-8 -*-

import sqlalchemy as sqa
from sqlalchemy.dialects import postgresql

from django.test import SimpleTestCase

from rest_framework import fields

from rest_enumfield import EnumField

from rest_witchcraft import field_mapping
from rest_witchcraft.fields import CharMappingField

from .models import Owner


class TestModelViewName(SimpleTestCase):
    def test_get_detail_view_name(self):

        name = field_mapping.get_detail_view_name(EnumField)

        self.assertEqual(name, "enumfield-detail")


class TestGetUrlKwargs(SimpleTestCase):
    def test_get_url_kwargs(self):
        kwargs = field_mapping.get_url_kwargs(Owner)
        self.assertDictEqual(
            {"read_only": True, "view_name": "owner-detail", "lookup_field": "id", "lookup_url_kwarg": "pk"}, kwargs
        )


class TestGetFieldType(SimpleTestCase):
    def test_get_field_type_can_map_string_column(self):

        field = field_mapping.get_field_type(sqa.Column(sqa.String()))

        self.assertTrue(issubclass(field, fields.CharField))

    def test_get_field_type_can_map_int_column(self):

        field = field_mapping.get_field_type(sqa.Column(sqa.BigInteger()))

        self.assertTrue(issubclass(field, fields.IntegerField))

    def test_get_field_type_can_map_float_column(self):

        field = field_mapping.get_field_type(sqa.Column(sqa.Float()))

        self.assertTrue(issubclass(field, fields.FloatField))

    def test_get_field_type_can_map_decimal_column(self):

        column = sqa.Column(sqa.Numeric(asdecimal=True))
        field = field_mapping.get_field_type(column)

        self.assertTrue(issubclass(field, fields.DecimalField))

    def test_get_field_type_can_map_interval_column(self):

        column = sqa.Column(sqa.Interval())
        field = field_mapping.get_field_type(column)

        self.assertTrue(issubclass(field, fields.DurationField))

    def test_get_field_type_can_map_time_column(self):

        column = sqa.Column(sqa.Time())
        field = field_mapping.get_field_type(column)

        self.assertTrue(issubclass(field, fields.TimeField))

    def test_get_field_type_can_map_datetime_column(self):

        column = sqa.Column(sqa.DateTime())
        field = field_mapping.get_field_type(column)

        self.assertTrue(issubclass(field, fields.DateTimeField))

    def test_get_field_type_can_map_date_column(self):

        column = sqa.Column(sqa.Date())
        field = field_mapping.get_field_type(column)

        self.assertTrue(issubclass(field, fields.DateField))

    def test_get_field_type_can_map_bool_column(self):

        column = sqa.Column(sqa.Boolean(), nullable=False)
        field = field_mapping.get_field_type(column)

        self.assertTrue(issubclass(field, fields.BooleanField))

    def test_get_field_type_can_map_nullable_bool_column(self):

        column = sqa.Column(sqa.Boolean(), nullable=True)
        field = field_mapping.get_field_type(column)

        self.assertTrue(issubclass(field, fields.NullBooleanField))

    def test_get_field_type_can_map_pg_hstore_column(self):

        column = sqa.Column(postgresql.HSTORE())
        field = field_mapping.get_field_type(column)

        self.assertTrue(issubclass(field, CharMappingField))

    def test_get_field_type_can_map_pg_array_column(self):

        column = sqa.Column(postgresql.ARRAY(item_type=sqa.Integer()))
        field = field_mapping.get_field_type(column)

        self.assertTrue(issubclass(field, fields.ListField))
        self.assertIsInstance(field().child, fields.IntegerField)

    def test_get_field_type_pg_array_column_raises_when_item_type_not_found(self):
        class DummyType(object):
            python_type = None

        column = sqa.Column(postgresql.ARRAY(item_type=DummyType))

        with self.assertRaises(KeyError):
            field_mapping.get_field_type(column)



# -*- coding: utf-8 -*-

from sqlalchemy import Column, create_engine, orm, types
from sqlalchemy.ext.declarative import declarative_base

from django.http import Http404
from django.test import SimpleTestCase

from rest_framework.test import APIRequestFactory

from rest_witchcraft import serializers, viewsets


factory = APIRequestFactory()

engine = create_engine("sqlite://")
session = orm.scoped_session(orm.sessionmaker(bind=engine))
Base = declarative_base()
Base.query = session.query_property()


class RouterTestModel(Base):
    __tablename__ = "routertest"
    id = Column(types.Integer(), default=3, primary_key=True)
    text = Column(types.String(length=200))


Base.metadata.create_all(engine)


class RouterTestModelSerializer(serializers.ModelSerializer):
    class Meta:
        model = RouterTestModel
        session = session
        fields = "__all__"


class TestModelRoutes(SimpleTestCase):
    def test_get_model_using_queryset(self):
        class RouterTestViewSet(viewsets.ModelViewSet):
            queryset = RouterTestModel.query
            serializer_class = RouterTestModelSerializer

        model = RouterTestViewSet.get_model()

        self.assertEqual(model, RouterTestModel)

    def test_get_model_using_serializer(self):
        class RouterTestViewSet(viewsets.ModelViewSet):
            serializer_class = RouterTestModelSerializer

        model = RouterTestViewSet.get_model()

        self.assertEqual(model, RouterTestModel)

    def test_get_model_fails_with_assert_error(self):
        class RouterTestViewSet(viewsets.ModelViewSet):
            pass

        with self.assertRaises(AssertionError):
            RouterTestViewSet.get_model()

    def test_get_object_raises_404(self):
        class RouterTestViewSet(viewsets.ModelViewSet):
            queryset = RouterTestModel.query
            serializer_class = RouterTestModelSerializer
            lookup_field = "id"
            lookup_url_kwarg = "pk"

        viewset = RouterTestViewSet()
        viewset.kwargs = {"pk": 1}

        with self.assertRaises(Http404):
            viewset.get_object()



# -*- coding: utf-8 -*-
import os


INSTALLED_APPS = ["rest_witchcraft", "rest_framework", "django.contrib.staticfiles"]

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_L10N = True

USE_TZ = True

USE_THOUSAND_SEPARATOR = True

SECRET_KEY = "secret"

ALLOWED_HOSTS = ["*"]

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [os.path.join(os.path.dirname(__file__), "templates")],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
            ]
        },
    }
]



# -*- coding: utf-8 -*-
"""
Provides generic filtering backends that can be used to filter the results
returned by list views.
"""


from sqlalchemy import func, or_
from sqlalchemy.sql import operators

from django.template import loader
from django.utils.encoding import force_text
from django.utils.translation import gettext_lazy

from rest_framework.compat import coreapi, coreschema
from rest_framework.filters import BaseFilterBackend
from rest_framework.settings import api_settings


class SearchFilter(BaseFilterBackend):
    search_param = api_settings.SEARCH_PARAM
    template = "rest_framework/filters/search.html"
    lookup_prefixes = {
        "": lambda c, x: operators.ilike_op(c, "%{0}%".format(x)),  # icontains
        "^": lambda c, x: c.ilike(x.replace("%", "%%") + "%"),  # istartswith
        "=": lambda c, x: func.lower(c) == func.lower(x),  # iequals
        "@": operators.eq,  # equals
    }
    search_title = gettext_lazy("Search")
    search_description = gettext_lazy("A search term.")

    def get_schema_fields(self, view):
        assert coreapi is not None, "coreapi must be installed to use `get_schema_fields()`"
        assert coreschema is not None, "coreschema must be installed to use `get_schema_fields()`"
        return [
            coreapi.Field(
                name=self.search_param,
                required=False,
                location="query",
                schema=coreschema.String(
                    title=force_text(self.search_title), description=force_text(self.search_description)
                ),
            )
        ]

    def get_schema_operation_parameters(self, view):
        return [
            {
                "name": self.search_param,
                "required": False,
                "in": "query",
                "description": force_text(self.search_description),
                "schema": {"type": "string"},
            }
        ]

    def get_search_fields(self, view, request):
        return getattr(view, "search_fields", None)

    def get_search_terms(self, request):
        params = request.query_params.get(self.search_param, "")
        params = params.replace("\x00", "")  # strip null characters
        params = params.replace(",", " ")
        return params.split()

    def to_html(self, request, queryset, view):
        if not getattr(view, "search_fields", None):
            return ""

        term = self.get_search_terms(request)
        term = term[0] if term else ""
        context = {"param": self.search_param, "term": term}
        template = loader.get_template(self.template)
        return template.render(context)

    def filter_queryset(self, request, queryset, view):
        search_fields = self.get_search_fields(view, request)
        search_terms = self.get_search_terms(request)

        if not search_fields or not search_terms:
            return queryset

        model = view.get_model()

        expressions = []
        for field in search_fields:
            for term in search_terms:
                expr = self.get_expression(model, field, term)
                if expr is not None:
                    expressions.append(expr)

        return queryset.filter(or_(*expressions))

    def get_expression(self, model, field, term):
        op = self.lookup_prefixes[""]
        if field[0] in self.lookup_prefixes:
            op = self.lookup_prefixes[field[0]]
            field = field[1:]

        expr = op(getattr(model, field), term)
        return expr



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import shutil
import sys

from setuptools import find_packages, setup


here = os.path.abspath(os.path.dirname(__file__))

about = {}
with open(os.path.join(here, "rest_witchcraft", "__version__.py")) as f:
    exec(f.read(), about)  # yapf: disable


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname), "rb").read().decode("utf-8")


if sys.argv[-1] == "publish":
    twine = shutil.which("twine")
    if twine is None:
        print("twine not installed.\nUse `pip install twine`.\nExiting.")
        sys.exit()
    os.system("python setup.py sdist bdist_wheel")
    os.system("twine upload dist/*")
    print("You probably want to also tag the version now:")
    print("  git tag -a %s -m %s" % (about["__version__"], about["__version__"]))
    print("  git push --tags")
    os.system("make clean")
    sys.exit()

setup(
    author=about["__author__"],
    author_email=about["__author_email__"],
    description=about["__description__"],
    install_requires=["djangorestframework", "six", "django-sorcery>=0.10.1", "django-rest-enumfield"],
    license="MIT",
    long_description=read("README.rst"),
    name="django-rest-witchcraft",
    packages=find_packages(exclude=["tests"]),
    url="https://github.com/shosca/django-rest-witchcraft",
    version=about["__version__"],
    keywords="sqlalchemy django rest framework drf rest_framework",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Framework :: Django :: 1.11",
        "Framework :: Django :: 2.0",
        "Framework :: Django :: 2.1",
        "Framework :: Django :: 2.2",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python",
        "Topic :: Internet :: WWW/HTTP",
    ],
)



# -*- coding: utf-8 -*-
from itertools import chain

import six

from sqlalchemy import orm

from django.db.models.constants import LOOKUP_SEP

from django_sorcery.db import meta

from rest_framework import mixins


class DestroyModelMixin(mixins.DestroyModelMixin):
    """
    Deletes a model instance
    """

    def perform_destroy(self, instance):
        session = self.get_session()
        session.delete(instance)


class QuerySerializerMixin(object):
    """
    Adds query serializer validation logic to viewset

    Query will be validated as part of query viewset initialization
    therefore query will be validated before any of the viewset actions
    are executed.

    In addition query serializer will be included in serializer context
    for standard viewset serializers. That
    """

    query_serializer_class = None

    @property
    def query_serializer(self):
        return getattr(self, "_query_serializer", None)

    @query_serializer.setter
    def query_serializer(self, value):
        self._query_serializer = value

    def get_query_serializer_class(self):
        return (
            self.query_serializer_class
            or getattr(self.get_serializer_class()(), "get_query_serializer_class", lambda: None)()
        )

    def get_query_serializer_context(self):
        return self.get_serializer_context()

    def get_query_serializer(self, *args, **kwargs):
        serializer_class = kwargs.pop("serializer_class", None) or self.get_query_serializer_class()
        if serializer_class is None:
            return
        kwargs.setdefault("context", self.get_query_serializer_context())
        kwargs.setdefault("data", dict(self.request.GET.lists()))
        self.query_serializer = serializer = serializer_class(*args, **kwargs)
        serializer.is_valid()
        return serializer

    def check_query(self):
        serializer = self.get_query_serializer()
        if serializer is not None:
            serializer.is_valid(raise_exception=True)

    def initial(self, request, *args, **kwargs):
        super().initial(request, *args, **kwargs)
        self.check_query()


class ExpandableQuerySerializerMixin(QuerySerializerMixin):
    """
    Adds expandable query serializer validation logic to viewset
    as well as automatic eager load of expanded fields on the serializer.

    The query serializer is expected to be generated by
    :py:meth:`rest_witchcraft.serializers.ExpandableModelSerializer.get_query_serializer_class`.
    """

    def get_queryset(self):
        queryset = super().get_queryset()

        serializer = self.query_serializer
        if serializer is None:
            return queryset

        return self.expand_queryset(queryset, chain(*serializer.validated_data.values()))

    def expand_queryset(self, queryset, values):
        to_expand = []

        for value in values:
            to_load = []
            components = value.split(LOOKUP_SEP)

            model = queryset._only_entity_zero().class_
            for c in components:
                props = meta.model_info(model).relationships
                try:
                    field = getattr(model, c)
                    model = props[c].relationship._dependency_processor.mapper.class_
                except (KeyError, AttributeError):
                    to_load = []
                    break
                else:
                    to_load.append(field)

            if to_load:
                to_expand.append(to_load)

        if to_expand:
            queryset = queryset.options(
                *[six.moves.reduce(lambda a, b: a.joinedload(b), expand, orm) for expand in to_expand]
            )

        return queryset

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context["query_serializer"] = self.query_serializer
        return context



# -*- coding: utf-8 -*-

from django_sorcery.db import meta

from rest_framework import routers


class DefaultRouter(routers.DefaultRouter):
    def get_default_base_name(self, viewset):

        model = getattr(viewset, "get_model", lambda: None)()

        assert model is not None, (
            "`base_name` argument not specified, and could not automatically determine the name from the viewset, "
            "as either queryset is is missing or is not a sqlalchemy query, or the serializer_class is not a "
            "sqlalchemy model serializer"
        )

        return model.__name__.lower()

    # for backwards compatibility DRF<3.9
    get_default_basename = get_default_base_name

    def get_lookup_regex(self, viewset, lookup_prefix=""):
        """
        Given a viewset, return the portion of the url regex that is used to match against a single instance.

        Can be overwritten by providing a `lookup_url_regex` on the viewset.
        """

        lookup_url_regex = getattr(viewset, "lookup_url_regex", None)
        if lookup_url_regex:
            return lookup_url_regex

        model = getattr(viewset, "get_model", lambda: None)()
        if model:
            info = meta.model_info(model)
            base_regex = "(?P<{lookup_prefix}{lookup_url_kwarg}>{lookup_value})"

            lookup_keys = [getattr(viewset, "lookup_url_kwarg", None) or getattr(viewset, "lookup_field", None)]
            if not lookup_keys[0] or len(info.primary_keys) > 1:
                lookup_keys = list(info.primary_keys)

            regexes = []
            for key in lookup_keys:
                regexes.append(
                    base_regex.format(lookup_prefix=lookup_prefix, lookup_url_kwarg=key, lookup_value="[^/.]+")
                )

            return "/".join(regexes)

        return super().get_lookup_regex(viewset, lookup_prefix)



# -*- coding: utf-8 -*-
"""
Field mapping from SQLAlchemy type's to DRF fields
"""
# -*- coding: utf-8 -*-
import datetime
import decimal

import six

from sqlalchemy.dialects import postgresql
from sqlalchemy.sql import sqltypes

from django_sorcery.db import meta

from rest_framework import fields

from rest_enumfield import EnumField

from .fields import CharMappingField


def get_detail_view_name(model):
    """
    Given a model class, return the view name to use for URL relationships
    that rever to instances of the model.
    """
    # TODO split camel case
    return "{}-detail".format(model.__name__.lower())


def get_url_kwargs(model):
    """
    Gets kwargs for the UriField
    """
    info = meta.model_info(model)
    lookup_field = list(info.primary_keys.keys())[0]

    field_kwargs = {
        "read_only": True,
        "view_name": get_detail_view_name(model),
        "lookup_field": lookup_field,
        "lookup_url_kwarg": "pk",
    }

    return field_kwargs


SERIALIZER_FIELD_MAPPING = {
    # sqlalchemy types
    postgresql.HSTORE: CharMappingField,
    sqltypes.Enum: EnumField,
    # python types
    datetime.date: fields.DateField,
    datetime.datetime: fields.DateTimeField,
    datetime.time: fields.TimeField,
    datetime.timedelta: fields.DurationField,
    decimal.Decimal: fields.DecimalField,
    float: fields.FloatField,
    int: fields.IntegerField,
    str: fields.CharField,
}
SERIALIZER_FIELD_MAPPING[six.text_type] = fields.CharField
for stype in six.string_types:
    SERIALIZER_FIELD_MAPPING[stype] = fields.CharField

try:
    from sqlalchemy_utils import types

    SERIALIZER_FIELD_MAPPING[types.IPAddressType] = fields.IPAddressField
    SERIALIZER_FIELD_MAPPING[types.UUIDType] = fields.UUIDField
    SERIALIZER_FIELD_MAPPING[types.URLType] = fields.URLField
except ImportError:  # pragma: no cover
    pass


def get_field_type(column):
    """
    Returns the field type to be used determined by the sqlalchemy column type or the column type's python type
    """
    if isinstance(column.type, sqltypes.Enum) and not column.type.enum_class:
        return fields.ChoiceField

    if isinstance(column.type, postgresql.ARRAY):
        child_field = SERIALIZER_FIELD_MAPPING.get(column.type.item_type.__class__) or SERIALIZER_FIELD_MAPPING.get(
            column.type.item_type.python_type
        )

        if child_field is None:
            raise KeyError("Could not figure out field for ARRAY item type '{}'".format(column.type.__class__))

        class ArrayField(fields.ListField):
            """Nested array field for PostreSQL's ARRAY type"""

            def __init__(self, *args, **kwargs):
                kwargs["child"] = child_field()
                super().__init__(*args, **kwargs)

        return ArrayField

    if column.type.__class__ in SERIALIZER_FIELD_MAPPING:
        return SERIALIZER_FIELD_MAPPING.get(column.type.__class__)

    if issubclass(column.type.python_type, bool):
        return fields.NullBooleanField if column.nullable else fields.BooleanField

    for typ in column.type.python_type.mro():
        if typ in SERIALIZER_FIELD_MAPPING:
            return SERIALIZER_FIELD_MAPPING.get(typ)
